<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace PrometheusModule\Controller;


use App\Model\Table\ContainersTable;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\AngularJS\Api;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use PrometheusModule\Model\Table\PrometheusExportersTable;

/**
 * Class PrometheusExportersController
 * @package PrometheusModule\Controller
 */
class PrometheusExportersController extends AppController {

    public function index() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }

        /** @var PrometheusExportersTable $PrometheusExportersTable */
        $PrometheusExportersTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusExporters');
        /** @var ContainersTable $ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like'   => [
                'PrometheusExporters.name',
                'PrometheusExporters.service',
            ],
            'equals' => [
                'PrometheusExporters.scrape_interval',
            ],
        ]);

        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        $all_exporters_tmp = $PrometheusExportersTable->getExportersIndex($GenericFilter, $PaginateOMat, $MY_RIGHTS);
        $all_exporters = [];
        foreach ($all_exporters_tmp as $exporter) {
            $exporter = $exporter->toArray();
            $exporter['container'] = '/' . $ContainersTable->treePath($exporter['container_id']);
            if ($this->hasRootPrivileges) {
                $exporter['allow_edit'] = true;
            } else {
                $exporter['allow_edit'] = $this->isWritableContainer($exporter['container_id']);
            }
            $all_exporters[] = $exporter;
        }

        $this->set('all_exporters', $all_exporters);
        $this->viewBuilder()->setOption('serialize', ['all_exporters']);

    }

    public function add() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }


        if ($this->request->is('post') || $this->request->is('put')) {
            /** @var PrometheusExportersTable $PrometheusExportersTable */
            $PrometheusExportersTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusExporters');

            $exporter = $PrometheusExportersTable->newEntity($this->request->getData('PrometheusExporter', []));
            $PrometheusExportersTable->save($exporter);
            if ($exporter->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $exporter->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            $this->set('PrometheusExporter', $exporter);
            $this->viewBuilder()->setOption('serialize', ['PrometheusExporter']);
        }

    }

    /**
     * @param int|null $id
     */
    public function edit($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var PrometheusExportersTable $PrometheusExportersTable */
        $PrometheusExportersTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusExporters');

        if (!$PrometheusExportersTable->existsById($id)) {
            throw new NotFoundException(__('Prometheus exporter not found'));
        }

        $exporter = $PrometheusExportersTable->get($id);

        if (!$this->allowedByContainerId($exporter->get('container_id'))) {
            $this->render403();
            return;
        }

        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            //Return user information
            $this->set('PrometheusExporter', $exporter);
            $this->viewBuilder()->setOption('serialize', ['PrometheusExporter']);
            return;
        }

        if ($this->request->is('post') || $this->request->is('put')) {
            $exporter->setAccess('id', false);

            $exporter = $PrometheusExportersTable->patchEntity($exporter, $this->request->getData('PrometheusExporter', []));
            $PrometheusExportersTable->save($exporter);
            if ($exporter->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $exporter->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            $this->set('PrometheusExporter', $exporter);
            $this->viewBuilder()->setOption('serialize', ['PrometheusExporter']);
        }
    }

    /**
     * @param int|null $id
     * @todo Add checks if the exporter is used or not before delete
     */
    public function delete($id = null) {
        if (!$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        /** @var PrometheusExportersTable $PrometheusExportersTable */
        $PrometheusExportersTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusExporters');

        if (!$PrometheusExportersTable->existsById($id)) {
            throw new NotFoundException(__('Prometheus exporter not found'));
        }

        $exporter = $PrometheusExportersTable->get($id);

        if (!$this->allowedByContainerId($exporter->get('container_id'))) {
            $this->render403();
            return;
        }

        if ($PrometheusExportersTable->delete($exporter)) {
            $this->set('success', true);
            $this->set('message', __('Prometheus exporter successfully deleted'));
            $this->viewBuilder()->setOption('serialize', ['success', 'message']);
            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->set('message', __('Error while deleting Prometheus exporter'));
        $this->viewBuilder()->setOption('serialize', ['success', 'message']);
    }

    /****************************
     *       AJAX METHODS       *
     ****************************/

    public function loadContainers() {
        if (!$this->isAngularJsRequest()) {
            throw new MethodNotAllowedException();
        }

        /** @var ContainersTable $ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');
        if ($this->hasRootPrivileges === true) {
            $containers = $ContainersTable->easyPath($this->MY_RIGHTS, OBJECT_HOST, [], $this->hasRootPrivileges);
        } else {
            $containers = $ContainersTable->easyPath($this->getWriteContainers(), OBJECT_HOST, [], $this->hasRootPrivileges);
        }
        $containers = Api::makeItJavaScriptAble($containers);

        $this->set('containers', $containers);
        $this->viewBuilder()->setOption('serialize', ['containers']);
    }

}
