<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace PrometheusModule\Lib;


use Cake\Log\Log;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Psr7\Request;
use itnovum\openITCOCKPIT\Core\FileDebugger;

class PrometheusApi {

    /**
     * @var string
     */
    private $address = '127.0.0.1';

    /**
     * @var int
     */
    private $port = 9090;

    /**
     * @var bool
     */
    private $useHttps = false;

    /**
     * @var Client|null
     */
    private $Client = null;

    /**
     * PrometheusApi constructor.
     * @param string $address
     * @param int $port
     */
    public function __construct(string $address = '127.0.0.1', int $port = 9090) {
        $this->address = $address;
        $this->port = $port;
    }

    public function switchToVicotriaMetrics() {
        $this->port = 8428;
        $this->Client = null;
        $this->Client = $this->getClient();
    }

    /**
     * @return Client
     */
    private function getClient() {
        if ($this->Client instanceof Client) {
            return $this->Client;
        }

        $options = [
            'http'    => false,
            'https'   => false,
            'verify'  => false,
            'headers' => [
                'Accept' => 'application/json'
            ]
        ];
        return new Client($options);
    }

    /**
     * @return string
     */
    private function getBaseUrl() {
        $protocol = 'http';
        if ($this->useHttps) {
            $protocol = 'https';
        }

        return sprintf(
            '%s://%s:%s/api/v1/',
            $protocol,
            $this->address,
            $this->port
        );
    }


    /**
     * @param string $hostUuid
     * @param array $filter
     * @param array $selected
     * @return array|mixed
     *
     * Filter Options
     * PrometheusQuery.name
     */
    public function queryingTargetMetadata(string $hostUuid, array $filter, array $selected) {
        $Client = $this->getClient();
        $url = $this->getBaseUrl();

        try {
            $result = $Client->request(
                'GET',
                $url . 'targets/metadata?match_target={uuid="' . $hostUuid . '"}'
            );
            //FileDebugger::dump($url . 'targets/metadata?match_target={uuid="' . $hostUuid . '"}');
        } catch (\Exception $e) {
            Log::error($e->getMessage());
            return [];
        }

        if ($result->getStatusCode() !== 200) {
            return [];
        }

        $resultArray = json_decode($result->getBody()->getContents(), true);
        if (!empty($filter['PrometheusQuery.name']) && isset($resultArray['data'])) {
            foreach ($resultArray['data'] as $index => $metric) {
                if (in_array($metric['metric'], $selected, true)) {
                    continue;
                }
                if (!preg_match('/' . $filter['PrometheusQuery.name'] . '/', $metric['metric'])) {
                    unset($resultArray['data'][$index]);
                }
            }
        }

        foreach ($resultArray['data'] as $index => $metric) {
            if (empty($metric['unit'])) {
                $resultArray['data'][$index]['unit'] = $this->getUnitByMetricName($metric['metric']);
            }
        }

        return $resultArray;
    }

    /**
     * @param string $hostUuid
     * @param string $metric
     * @param int $startTimestamp
     * @param int $endTimestamp
     * @param float $step
     * @return array|mixed
     */
    public function getPerfdataByUuidAndMetric(string $hostUuid, string $metric, int $startTimestamp, int $endTimestamp, ?float $step = null) {

        $Client = $this->getClient();
        $url = $this->getBaseUrl();

        if ($step === null) {
            $step = $this->getStepSize($startTimestamp, $endTimestamp);
        }

        $url = sprintf(
        //'%squery_range?query=%s&start=%s&end=%s&step=15s&match_target={uuid=\"' . $hostUuid . '\"}',
        //'%sseries?match[]=%s{uuid="%s"}',
            '%squery_range?query=%s{uuid="%s"}&start=%s&end=%s&step=%F',
            $url,
            $metric,
            $hostUuid,
            $startTimestamp,
            $endTimestamp,
            $step
        );

        try {
            $result = $Client->request(
                'GET',
                $url
            );
        } catch (\Exception $e) {
            Log::error($e->getMessage());
            return [];
        }

        if ($result->getStatusCode() !== 200) {
            return [];
        }

        return json_decode($result->getBody()->getContents(), true);
    }

    /**
     * @param string $promQl
     * @param int $startTimestamp
     * @param int $endTimestamp
     * @param null|float $step
     * @return array|mixed
     */
    public function getPerfdataByPromQL(string $promQl, int $startTimestamp, int $endTimestamp, ?float $step = null) {
        $Client = $this->getClient();
        $url = $this->getBaseUrl();

        if ($step === null) {
            $step = $this->getStepSize($startTimestamp, $endTimestamp);
        }

        $url = sprintf(
            '%squery_range?query=%s&start=%s&end=%s&step=%F',
            $url,
            $promQl,
            $startTimestamp,
            $endTimestamp,
            $step
        );

        try {
            $result = $Client->request(
                'GET',
                $url
            );
        } catch (\Exception $e) {
            Log::error($e->getMessage());
            return [];
        }

        if ($result->getStatusCode() !== 200) {
            return [];
        }

        return json_decode($result->getBody()->getContents(), true);
    }

    /**
     * @param string $metric
     * @return string
     */
    public function getUnitByMetricName(string $metric): string {
        // Try to get unit from metric name
        // https://prometheus.io/docs/practices/naming/#base-units
        $units = ['seconds', 'celsius', 'meters', 'bytes', 'ratio', 'volts', 'amperes', 'joules', 'grams', 'mhz'];
        $metricUnit = ''; //Unknown unit
        foreach ($units as $unit) {
            if (preg_match('/_' . $unit . '/', $metric)) {
                $metricUnit = $unit;
                if ($unit === 'ratio') {
                    $metricUnit = '%';
                }
            }
        }

        return $metricUnit;
    }

    /**
     * @param string $hostUuid
     * @param string $metric
     * @return array
     */
    public function getCurrentValueByMetric(string $hostUuid, string $metric) {
        $Client = $this->getClient();
        $url = $this->getBaseUrl();

        try {
            $result = $Client->request(
                'GET',
                $url . 'query?query=' . $metric . '{uuid="' . $hostUuid . '"}'
            );
        } catch (\Exception $e) {
            Log::error($e->getMessage());
            return [];
        }

        if ($result->getStatusCode() !== 200) {
            return [];
        }

        $resultArray = json_decode($result->getBody()->getContents(), true);
        return $resultArray;
    }

    /**
     * @param string $metric
     * @return array
     * @throws ClientException
     */
    public function getValueByMetric(string $metric) {
        $Client = $this->getClient();
        $url = $this->getBaseUrl();
        try {
            $result = $Client->request(
                'GET',
                $url . 'query?query=' . rawurlencode($metric),
                ['decode_content' => false]
            );
        } catch (ClientException $e) {
            if ($e->getCode() === 400) {
                //Bad request -> Query Error
                throw $e;
            }

            //Log error and return empty array
            Log::error($e->getMessage());
            return [];
        } catch (\Exception $e) {
            Log::error($e->getMessage());
            return [];
        }

        if ($result->getStatusCode() !== 200) {
            return [];
        }

        $resultArray = json_decode($result->getBody()->getContents(), true);
        return $resultArray;
    }

    /**
     * @return array|mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function getPendingAndFiringAlerts() {
        $Client = $this->getClient();
        $url = $this->getBaseUrl();

        try {
            $result = $Client->request(
                'GET',
                $url . 'alerts'
            );
            //FileDebugger::dump($url . 'targets/metadata?match_target={uuid="' . $hostUuid . '"}');
        } catch (\Exception $e) {
            Log::error($e->getMessage());
            return [];
        }

        if ($result->getStatusCode() !== 200) {
            return [];
        }

        $resultArray = json_decode($result->getBody()->getContents(), true);
        if (!isset($resultArray['status']) && $resultArray['status'] !== 'success') {
            return [];
        }

        return $resultArray;
    }

    /**
     * @param int $startTimestamp
     * @param int $endTimestamp
     * @return float
     */
    private function getStepSize(int $startTimestamp, int $endTimestamp) {
        // Auto calculate the step size for given time range
        // This highly inspired from Prometheus itself - many thanks
        // https://github.com/prometheus/prometheus/blob/49f775d8a0d76532fae36c0392d929bbf110577a/web/ui/react-app/src/pages/graph/Panel.tsx#L128-L142
        // this.props.options.range == selected time range in milliseconds (1 week, 2 days, etc)
        return max(floor(($endTimestamp - $startTimestamp) / 250), 1);
    }
}
