<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace PrometheusModule\Lib;


use Cake\Log\Log;
use Cake\ORM\TableRegistry;
use GuzzleHttp\Exception\ClientException;
use itnovum\openITCOCKPIT\Core\Views\Service;
use itnovum\openITCOCKPIT\Perfdata\UnitScaler;
use PrometheusModule\Model\Table\PrometheusAlertRulesTable;

class PrometheusPerfdataLoader {

    /**
     * @param Service $Service
     * @param int $startTimestamp
     * @param int $endTimestamp
     * @param bool $jsTimestamp
     * @param bool $scale
     * @param string|null $forcedUnit
     * @param bool $debug
     * @param string|null $gauge
     * @return array
     */
    public function getPerfdataByUuid(Service $Service, int $startTimestamp, int $endTimestamp, $jsTimestamp = false, $scale = true, $forcedUnit = null, $debug = false, ?string $gauge = null) {
        $serviceName = $Service->getServicename();

        $PrometheusApi = new PrometheusApi();
        // Query VictoriaMetrics instead of Prometheus
        // VictoriaMetrics is a long-term storage for Prometheus
        /*******************************************************
         *                    FAIR WARNING                     *
         *                                                     *
         * READ THIS IF YOU RECEIVE "WRONG" AMOUNTS OF METRICS *
         *                                                     *
         * If you are receiving a different metrics count      *
         * than Prometheus reports, for example if you         *
         * execute a PromQL like count(node_network_up)        *
         * Prometheus result: 40                               *
         * VictoriaMetrics result: 138                         *
         * than this is most likely because VictoriaMetrics    *
         * is storing the data for multiple months versus      *
         * Prometheus which is storing the data for a few days *
         *******************************************************/
        $PrometheusApi->switchToVicotriaMetrics();

        /** @var PrometheusAlertRulesTable $PrometheusAlertRulesTable */
        $PrometheusAlertRulesTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusAlertRules');
        $alertRule = $PrometheusAlertRulesTable->getAlertRuleByServiceId($Service->getId());

        if ($gauge === null || $gauge === '') {
            // Load all available metrics/gauges of the alert rule
            $tmpPerfData = $PrometheusApi->getPerfdataByPromQL($alertRule->get('promql'), $startTimestamp, $endTimestamp);
        } else {
            if (is_numeric($gauge)) {
                // If gauge is numeric (0,1,2,...) Prometheus executed an range(), sum() or whatever query, which returns multiple values but has no unique name because the __name__ key is missing in the result.
                // In this case, openITCOCKPIT displays the first gauge (0), second gauge (1) etc from the result array.
                $allTmpPerfData = $PrometheusApi->getPerfdataByPromQL($alertRule->get('promql'), $startTimestamp, $endTimestamp);
                $tmpPerfData = [
                    'status' => $allTmpPerfData['status'] ?? 'error',
                    'data'   => [
                        'resultType' => '???',
                        'result'     => [],
                    ]
                ];
                if (isset($allTmpPerfData['data']['result'][$gauge])) {
                    $tmpPerfData['data']['resultType'] = $allTmpPerfData['data']['resultType'];
                    $tmpPerfData['data']['result'][0] = $allTmpPerfData['data']['result'][$gauge];
                }
            } else {
                // PromQL passed via URL. Only load the specific gauge/metric
                $tmpPerfData = $PrometheusApi->getPerfdataByPromQL($gauge, $startTimestamp, $endTimestamp);
            }
        }

        $performance_data = [];

        if (isset($tmpPerfData['status']) && isset($tmpPerfData['data'])) {
            if ($tmpPerfData['status'] === "success") {

                // One PromQL can return multiple metrics but all metrics will have the same thresholds
                $warning = $critical = $min = $max = null;
                if ($alertRule->threshold_type === 'scalar') {
                    $warning = $alertRule->warning_min;
                    $critical = $alertRule->critical_min;

                    if ($alertRule->unit === '%') {
                        $min = 0;
                        $max = 100;
                    }
                }

                if (isset($tmpPerfData['data']['result'][0]['values'])) {
                    $doesPromQlReturnMoreThanOneMetric = sizeof($tmpPerfData['data']['result']) > 1;

                    //Make it look like Nagios performance data...
                    foreach ($tmpPerfData['data']['result'] as $metric) {
                        $promql = $alertRule->get('promql'); // Single result metrics can use the PromQL itself
                        if($doesPromQlReturnMoreThanOneMetric) {
                            // Use the unique (or not unique) matric name from Prometheus
                            $promql = PromQl::getPromQlFromMetric($metric['metric']);
                        }

                        $datasource = [
                            'ds'         => $promql,
                            'name'       => $promql,
                            'label'      => $promql,
                            'unit'       => $alertRule->get('unit'),
                            'act'        => floatval($metric['values'][sizeof($metric['values']) - 1][1]),
                            'warn'       => $warning,
                            'crit'       => $critical,
                            'min'        => $min,
                            'max'        => $max,
                            'metric'     => $promql,
                            'prometheus' => $metric['metric']
                        ];

                        $data = [];
                        foreach ($metric['values'] as $record) {
                            if ($debug === true) {
                                $data[date('d.m.Y H:i:s', $record[0])] = floatval($record[1]);
                            } else {
                                if ($jsTimestamp) {
                                    $data[$record[0] * 1000] = floatval($record[1]);
                                } else {
                                    $data[$record[0]] = floatval($record[1]);
                                }
                            }
                        }

                        $performance_data[] = [
                            'datasource' => $datasource,
                            'data'       => $data
                        ];

                    }
                }
            }
        }

        if ($scale === true) {
            foreach ($performance_data as $index => $gauge) {
                $UnitScaler = new UnitScaler($gauge);

                if ($forcedUnit === null) {
                    $performance_data[$index] = $UnitScaler->scale();
                } else {
                    $performance_data[$index] = $UnitScaler->scaleToUnit($forcedUnit);
                }
            }
        }

        return $performance_data;
    }

    /**
     * Query Prometheus and returns an array with all available Prometheus Metrics for a given Service ID
     * This is used by services/browser
     * @param Service $Service
     * @param bool $includePromQl Include the base PromQL in the result - this is used by Grafana to display all metrics from the PromQL ("value" in legacy terms)
     * @param bool $applyNumericKeyWorkaround Used for Service Browser or Maps to select a Metric from non unique rate() or sum() prometheus results
     * @return array
     */
    public function getAvailableMetricsByService(Service $Service, bool $includePromQl = false, bool $applyNumericKeyWorkaround = false): array {
        $serviceName = $Service->getServicename();

        $PrometheusApi = new PrometheusApi();

        /** @var PrometheusAlertRulesTable $PrometheusAlertRulesTable */
        $PrometheusAlertRulesTable = TableRegistry::getTableLocator()->get('PrometheusModule.PrometheusAlertRules');
        $alertRule = $PrometheusAlertRulesTable->getAlertRuleByServiceId($Service->getId());

        $perfdata = [];

        try {
            $metricValue = $PrometheusApi->getValueByMetric($alertRule['promql']);
            if (isset($metricValue['status']) && isset($metricValue['data'])) {
                if ($metricValue['status'] === "success") {

                    // One PromQL can return multiple metrics but all metrics will have the same thresholds
                    $warning = $critical = $min = $max = null;
                    if ($alertRule->threshold_type === 'scalar') {
                        $warning = $alertRule->warning_min;
                        $critical = $alertRule->critical_min;

                        if ($alertRule->unit === '%') {
                            $min = 0;
                            $max = 100;
                        }
                    }

                    // Some tools (Grafana for example) can handle PromQL with multiple metrics by itself.
                    // For this reason we include the PromQL into the result set (if multiple metrics are available and $includePromQl === true)
                    if (sizeof($metricValue['data']['result']) > 1 && $includePromQl === true) {
                        $promql = $alertRule->get('promql');
                        $perfdata[$promql] = [
                            'current'  => null, // The PromQL metric itself has no value
                            'unit'     => $alertRule->get('unit'),
                            'warning'  => $warning,
                            'critical' => $critical,
                            'min'      => $min,
                            'max'      => $max,

                            // The label key is for compatibility with User Defined Grafana Dashboards.
                            // Before ITC-2824 openITCOCKPIT stored "value" as datasource for Prometheus services in
                            // the DB table grafana_userdashboard_metrics. This value was not used because
                            // openITCOCKPIT syncs automatically the defined PromQL from the Alert Rule to Grafana.
                            // This label is used, to be backwards compatible
                            'metric'   => 'value'
                        ];
                    }

                    $doesPromQlReturnMoreThanOneMetric = sizeof($metricValue['data']['result']) > 1;
                    $i = 0;
                    $uniqueMetrics = true;
                    foreach ($metricValue['data']['result'] as $metric) {
                        // Use the PromQL from the alert rule
                        $promql = $alertRule->promql;

                        if ($doesPromQlReturnMoreThanOneMetric === true) {
                            // The PromQL from the Alert Rule returns more than one metric
                            // Build a new PromQL to query one single metric
                            if (isset($metric['metric']['__name__'])) {
                                // Only rebuild a unique PromQL for unique metrics
                                // If Prometheus does not return the __name__ key, this metric is not unique (rate() or sum() or other function)
                                $promql = PromQl::getPromQlFromMetric($metric['metric']);
                            } else {
                                if ($applyNumericKeyWorkaround) {
                                    // Use the none unique PromQL as displayName for the metric
                                    $promql = PromQl::getPromQlFromMetric($metric['metric']);
                                }
                                $uniqueMetrics = false;
                            }
                        }

                        $key = $promql;
                        if (!isset($metric['metric']['__name__']) && $applyNumericKeyWorkaround === true) {
                            // Queries with function (like sum(node_load1), rate(node_load1[1m]), etc...) have no unique key
                            // This is a hacky workaround used by /services/browser for example
                            $key = $i;
                            $uniqueMetrics = false;
                        }

                        $perfdata[$key] = [
                            'current'    => floatval($metric['value'][1]),
                            'unit'       => $alertRule->get('unit'),
                            'warning'    => $warning,
                            'critical'   => $critical,
                            'min'        => $min,
                            'max'        => $max,
                            'metric'     => $doesPromQlReturnMoreThanOneMetric ? $promql : 'value',
                            'prometheus' => $metric['metric']
                        ];
                        $i++;

                        if ($applyNumericKeyWorkaround === false && $uniqueMetrics === false) {
                            // This PromQL returns multiple values but has no unique key for each metric
                            // Due to $applyNumericKeyWorkaround is disabled, we return only the given PromQL itself as available metric
                            // This is for Grafana for example
                            // Used by rate() and sum() and other functions
                            break;
                        }

                    }
                }
            }
        } catch (ClientException $e) {
            // Error while loading data from Prometheus
            Log::error($e->getMessage());
        }

        return $perfdata;
    }

    /**
     * @param string $promql
     * @return int
     */
    public function getMetricsCountByPromQl(string $promql): int {
        $PrometheusApi = new PrometheusApi();

        try {
            $metricValue = $PrometheusApi->getValueByMetric($promql);
            if (isset($metricValue['status']) && isset($metricValue['data'])) {
                if ($metricValue['status'] === "success") {
                    return sizeof($metricValue['data']['result']);
                }
            }
        } catch (ClientException $e) {
            // Error while loading data from Prometheus
            Log::error($e->getMessage());
        }

        return 1;
    }

}
